import logging

from pprintpp import pformat

__all__ = ["logger", "log_state"]

# create logger
logger = logging.getLogger("PyCardano")

# create console handler and set level to debug
ch = logging.StreamHandler()

# create formatter
formatter = logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")

# add formatter to ch
ch.setFormatter(formatter)

# add ch to logger
logger.addHandler(ch)


def log_state(func):
    """Decorator to log the state of an object after its function call."""

    def wrapper(obj, *args, **kwargs):
        try:
            output = func(obj, *args, **kwargs)
            logger.debug(
                f"Class: {obj.__class__}, method: {func}, state:\n {pformat(vars(obj), indent=2)}"
            )
            return output
        except Exception as e:
            logger.warning(
                f"Class: {obj.__class__}, method: {func}, state:\n {pformat(vars(obj), indent=2)}"
            )
            raise e

    return wrapper



"""A module that contains address-related classes.

Specifications and references could be found in:
    - CIP-0005: https://github.com/cardano-foundation/CIPs/tree/master/CIP-0005
    - CIP-0019: https://github.com/cardano-foundation/CIPs/tree/master/CIP-0019

"""

from __future__ import annotations

import binascii
import os
from enum import Enum
from typing import Optional, Type, Union

import base58
from cbor2 import CBORTag
from typing_extensions import override

from pycardano.cbor import cbor2
from pycardano.crypto.bech32 import decode, encode
from pycardano.exception import (
    DecodingException,
    DeserializeException,
    InvalidAddressInputException,
)
from pycardano.hash import VERIFICATION_KEY_HASH_SIZE, ScriptHash, VerificationKeyHash
from pycardano.network import Network
from pycardano.serialization import CBORSerializable, limit_primitive_type

__all__ = ["AddressType", "PointerAddress", "Address"]


class AddressType(Enum):
    """
    Address type definition.
    """

    BYRON = 0b1000
    """Byron address"""

    KEY_KEY = 0b0000
    """Payment key hash + Stake key hash"""

    SCRIPT_KEY = 0b0001
    """Script hash + Stake key hash"""

    KEY_SCRIPT = 0b0010
    """Payment key hash + Script hash"""

    SCRIPT_SCRIPT = 0b0011
    """Script hash + Script hash"""

    KEY_POINTER = 0b0100
    """Payment key hash + Pointer address"""

    SCRIPT_POINTER = 0b0101
    """Script hash + Pointer address"""

    KEY_NONE = 0b0110
    """Payment key hash only"""

    SCRIPT_NONE = 0b0111
    """Script hash for payment part only"""

    NONE_KEY = 0b1110
    """Stake key hash for stake part only"""

    NONE_SCRIPT = 0b1111
    """Script hash for stake part only"""


class PointerAddress(CBORSerializable):
    """Pointer address.

    It refers to a point of the chain containing a stake key registration certificate.

    Args:
        slot (int): Slot in which the staking certificate was posted.
        tx_index (int): The transaction index (within that slot).
        cert_index (int): A (delegation) certificate index (within that transaction).
    """

    def __init__(self, slot: int, tx_index: int, cert_index: int):
        self._slot = slot
        self._tx_index = tx_index
        self._cert_index = cert_index

    @property
    def slot(self) -> int:
        return self._slot

    @property
    def tx_index(self) -> int:
        return self._tx_index

    @property
    def cert_index(self) -> int:
        return self._cert_index

    def encode(self) -> bytes:
        """Encode the pointer address to bytes.

        The encoding follows
        `CIP-0019#Pointers <https://github.com/cardano-foundation/CIPs/tree/master/CIP-0019#pointers>`_.

        Returns:
            bytes: Encoded bytes.

        Examples:
            >>> PointerAddress(1, 2, 3).encode()
            b'\\x01\\x02\\x03'
            >>> PointerAddress(123456789, 2, 3).encode()
            b'\\xba\\xef\\x9a\\x15\\x02\\x03'
        """

        def _encode_int(n):
            output = bytearray()
            output.append(n & 0x7F)
            n >>= 7
            while n > 0:
                output.append(0x80 | (n & 0x7F))
                n >>= 7
            output.reverse()
            return bytes(output)

        return (
            _encode_int(self.slot)
            + _encode_int(self.tx_index)
            + _encode_int(self.cert_index)
        )

    @classmethod
    def decode(cls, data: bytes) -> PointerAddress:
        """Decode bytes into a PointerAddress.

        Args:
            data (bytes): The data to be decoded.

        Returns:
            PointerAddress: Decoded pointer address.

        Examples:
            >>> PointerAddress.decode(b'\\x01\\x02\\x03')
            PointerAddress(1, 2, 3)
            >>> PointerAddress.decode(b'\\xba\\xef\\x9a\\x15\\x02\\x03')
            PointerAddress(123456789, 2, 3)
        """
        ints = []
        cur_int = 0
        for i in data:
            cur_int |= i & 0x7F
            if not (i & 0x80):
                ints.append(cur_int)
                cur_int = 0
            else:
                cur_int <<= 7

        if len(ints) != 3:
            raise DecodingException(
                f"Error in decoding data {data} into a PointerAddress"
            )

        return cls(*ints)

    def to_primitive(self) -> bytes:
        return self.encode()

    @classmethod
    @limit_primitive_type(bytes)
    def from_primitive(cls: Type[PointerAddress], value: bytes) -> PointerAddress:
        return cls.decode(value)

    def __eq__(self, other):
        if not isinstance(other, PointerAddress):
            return False
        else:
            return (
                other.slot == self.slot
                and other.tx_index == self.tx_index
                and other.cert_index == self.cert_index
            )

    def __repr__(self):
        return f"PointerAddress({self.slot}, {self.tx_index}, {self.cert_index})"


class Address(CBORSerializable):
    """A shelley address. It consists of two parts: payment part and staking part.
        Either of the parts could be None, but they cannot be None at the same time.

    Args:
        payment_part (Union[VerificationKeyHash, ScriptHash, None]): Payment part of the address.
        staking_part (Union[KeyHash, ScriptHash, PointerAddress, None]): Staking part of the address.
        network (Network): Type of network the address belongs to.
    """

    def __init__(
        self,
        payment_part: Union[VerificationKeyHash, ScriptHash, None] = None,
        staking_part: Union[
            VerificationKeyHash, ScriptHash, PointerAddress, None
        ] = None,
        network: Network = Network.MAINNET,
    ):
        self._payment_part = payment_part
        self._staking_part = staking_part
        self._network = network

        # Byron address fields (only populated when decoding Byron addresses)
        self._byron_payload_hash: Optional[bytes] = None
        self._byron_attributes: Optional[dict] = None
        self._byron_type: Optional[int] = None
        self._byron_crc32: Optional[int] = None

        self._address_type = self._infer_address_type()
        self._header_byte = self._compute_header_byte() if not self.is_byron else None
        self._hrp = self._compute_hrp() if not self.is_byron else None

    @property
    def is_byron(self) -> bool:
        """Check if this is a Byron-era address.

        Returns:
            bool: True if this is a Byron address, False if Shelley/later.
        """
        return self._byron_payload_hash is not None

    def _infer_address_type(self):
        """Guess address type from the combination of payment part and staking part."""
        # Check if this is a Byron address
        if self.is_byron:
            return AddressType.BYRON

        payment_type = type(self.payment_part)
        staking_type = type(self.staking_part)
        if payment_type == VerificationKeyHash:
            if staking_type == VerificationKeyHash:
                return AddressType.KEY_KEY
            elif staking_type == ScriptHash:
                return AddressType.KEY_SCRIPT
            elif staking_type == PointerAddress:
                return AddressType.KEY_POINTER
            elif self.staking_part is None:
                return AddressType.KEY_NONE
        elif payment_type == ScriptHash:
            if staking_type == VerificationKeyHash:
                return AddressType.SCRIPT_KEY
            elif staking_type == ScriptHash:
                return AddressType.SCRIPT_SCRIPT
            elif staking_type == PointerAddress:
                return AddressType.SCRIPT_POINTER
            elif self.staking_part is None:
                return AddressType.SCRIPT_NONE
        elif self.payment_part is None:
            if staking_type == VerificationKeyHash:
                return AddressType.NONE_KEY
            elif staking_type == ScriptHash:
                return AddressType.NONE_SCRIPT

        raise InvalidAddressInputException(
            f"Cannot construct a shelley address from a combination of "
            f"payment part: {self.payment_part} and "
            f"stake part: {self.staking_part}"
        )

    @property
    def payment_part(self) -> Union[VerificationKeyHash, ScriptHash, None]:
        """Payment part of the address."""
        return self._payment_part

    @property
    def staking_part(
        self,
    ) -> Union[VerificationKeyHash, ScriptHash, PointerAddress, None]:
        """Staking part of the address."""
        return self._staking_part

    @property
    def network(self) -> Network:
        """Network this address belongs to."""
        return self._network

    @property
    def address_type(self) -> AddressType:
        """Address type."""
        return self._address_type

    @property
    def header_byte(self) -> Optional[bytes]:
        """Header byte that identifies the type of address. None for Byron addresses."""
        return self._header_byte

    @property
    def hrp(self) -> Optional[str]:
        """Human-readable prefix for bech32 encoder. None for Byron addresses."""
        return self._hrp

    @property
    def payload_hash(self) -> Optional[bytes]:
        """Byron address payload hash (28 bytes). None for Shelley addresses."""
        return self._byron_payload_hash if self.is_byron else None

    @property
    def byron_attributes(self) -> Optional[dict]:
        """Byron address attributes. None for Shelley addresses."""
        return self._byron_attributes if self.is_byron else None

    @property
    def byron_type(self) -> Optional[int]:
        """Byron address type (0=Public Key, 2=Redemption). None for Shelley addresses."""
        return self._byron_type if self.is_byron else None

    @property
    def crc32_checksum(self) -> Optional[int]:
        """Byron address CRC32 checksum. None for Shelley addresses."""
        return self._byron_crc32 if self.is_byron else None

    def _compute_header_byte(self) -> bytes:
        """Compute the header byte."""
        return (self.address_type.value << 4 | self.network.value).to_bytes(
            1, byteorder="big"
        )

    def _compute_hrp(self) -> str:
        """Compute human-readable prefix for bech32 encoder.

        Based on
        `miscellaneous section <https://github.com/cardano-foundation/CIPs/tree/master/CIP-0005#miscellaneous>`_
        in CIP-5.
        """
        prefix = (
            "stake"
            if self.address_type in (AddressType.NONE_KEY, AddressType.NONE_SCRIPT)
            else "addr"
        )
        suffix = "" if self.network == Network.MAINNET else "_test"
        return prefix + suffix

    def __bytes__(self):
        if self.is_byron:
            payload = cbor2.dumps(
                [
                    self._byron_payload_hash,
                    self._byron_attributes,
                    self._byron_type,
                ]
            )
            return cbor2.dumps([CBORTag(24, payload), self._byron_crc32])

        payment = self.payment_part or bytes()
        if self.staking_part is None:
            staking = bytes()
        elif type(self.staking_part) is PointerAddress:
            staking = self.staking_part.encode()
        else:
            staking = self.staking_part
        return self.header_byte + bytes(payment) + bytes(staking)

    def encode(self) -> str:
        """Encode the address in Bech32 format (Shelley) or Base58 format (Byron).

        More info about Bech32 `here <https://github.com/bitcoin/bips/blob/master/bip-0173.mediawiki#Bech32>`_.

        Returns:
            str: Encoded address in Bech32 (Shelley) or Base58 (Byron).

        Examples:
            >>> payment_hash = VerificationKeyHash(
            ...     bytes.fromhex("cc30497f4ff962f4c1dca54cceefe39f86f1d7179668009f8eb71e59"))
            >>> print(Address(payment_hash).encode())
            addr1v8xrqjtlfluk9axpmjj5enh0uw0cduwhz7txsqyl36m3ukgqdsn8w
        """
        if self.is_byron:
            return base58.b58encode(bytes(self)).decode("ascii")
        return encode(self.hrp, bytes(self))

    @classmethod
    def decode(cls, data: str) -> Address:
        """Decode a bech32 string into an address object.

        Args:
            data (str): Bech32-encoded string.

        Returns:
            Address: Decoded address.

        Raises:
            DecodingException: When the input string is not a valid Shelley address.

        Examples:
            >>> addr = Address.decode("addr1v8xrqjtlfluk9axpmjj5enh0uw0cduwhz7txsqyl36m3ukgqdsn8w")
            >>> khash = VerificationKeyHash(bytes.fromhex("cc30497f4ff962f4c1dca54cceefe39f86f1d7179668009f8eb71e59"))
            >>> assert addr == Address(khash)
        """
        return cls.from_primitive(data)

    def to_primitive(self) -> bytes:
        return bytes(self)

    @classmethod
    @limit_primitive_type(bytes, str)
    def from_primitive(cls: Type[Address], value: Union[bytes, str]) -> Address:
        # Convert string to bytes
        if isinstance(value, str):
            # Check for Byron Base58 prefixes (common Byron patterns)
            if value.startswith(("Ae2td", "Ddz")):
                return cls._from_byron_base58(value)

            # Try Bech32 decode for Shelley addresses
            original_str = value
            try:
                value = bytes(decode(value))
            except Exception:
                try:
                    return cls._from_byron_base58(original_str)
                except Exception as e:
                    raise DecodingException(f"Failed to decode address string: {e}")

        # At this point, value is always bytes
        # Check if it's a Byron address (CBOR with tag 24)
        try:
            decoded = cbor2.loads(value)
            if isinstance(decoded, (tuple, list)) and len(decoded) == 2:
                if isinstance(decoded[0], CBORTag) and decoded[0].tag == 24:
                    # This is definitely a Byron address - validate and decode it
                    return cls._from_byron_cbor(value)
        except DecodingException:
            # Byron decoding failed with validation error - re-raise it
            raise
        except Exception:
            # Not Byron CBOR (general CBOR decode error), continue with Shelley decoding
            pass

        # Shelley address decoding (existing logic)
        header = value[0]
        payload = value[1:]
        addr_type = AddressType((header & 0xF0) >> 4)
        network = Network(header & 0x0F)
        if addr_type == AddressType.KEY_KEY:
            return cls(
                VerificationKeyHash(payload[:VERIFICATION_KEY_HASH_SIZE]),
                VerificationKeyHash(payload[VERIFICATION_KEY_HASH_SIZE:]),
                network,
            )
        elif addr_type == AddressType.KEY_SCRIPT:
            return cls(
                VerificationKeyHash(payload[:VERIFICATION_KEY_HASH_SIZE]),
                ScriptHash(payload[VERIFICATION_KEY_HASH_SIZE:]),
                network,
            )
        elif addr_type == AddressType.KEY_POINTER:
            pointer_addr = PointerAddress.decode(payload[VERIFICATION_KEY_HASH_SIZE:])
            return cls(
                VerificationKeyHash(payload[:VERIFICATION_KEY_HASH_SIZE]),
                pointer_addr,
                network,
            )
        elif addr_type == AddressType.KEY_NONE:
            return cls(VerificationKeyHash(payload), None, network)
        elif addr_type == AddressType.SCRIPT_KEY:
            return cls(
                ScriptHash(payload[:VERIFICATION_KEY_HASH_SIZE]),
                VerificationKeyHash(payload[VERIFICATION_KEY_HASH_SIZE:]),
                network,
            )
        elif addr_type == AddressType.SCRIPT_SCRIPT:
            return cls(
                ScriptHash(payload[:VERIFICATION_KEY_HASH_SIZE]),
                ScriptHash(payload[VERIFICATION_KEY_HASH_SIZE:]),
                network,
            )
        elif addr_type == AddressType.SCRIPT_POINTER:
            pointer_addr = PointerAddress.decode(payload[VERIFICATION_KEY_HASH_SIZE:])
            return cls(
                ScriptHash(payload[:VERIFICATION_KEY_HASH_SIZE]), pointer_addr, network
            )
        elif addr_type == AddressType.SCRIPT_NONE:
            return cls(ScriptHash(payload), None, network)
        elif addr_type == AddressType.NONE_KEY:
            return cls(None, VerificationKeyHash(payload), network)
        elif addr_type == AddressType.NONE_SCRIPT:
            return cls(None, ScriptHash(payload), network)
        raise DeserializeException(f"Error in deserializing bytes: {value}")

    @classmethod
    def _from_byron_base58(cls: Type[Address], base58_str: str) -> Address:
        """Decode a Byron address from Base58 string.

        Args:
            base58_str: Base58-encoded Byron address string.

        Returns:
            Address: Decoded Byron address instance.

        Raises:
            DecodingException: When decoding fails.
        """
        try:
            cbor_bytes = base58.b58decode(base58_str)
        except Exception as e:
            raise DecodingException(f"Failed to decode Base58 string: {e}")

        return cls._from_byron_cbor(cbor_bytes)

    @classmethod
    def _from_byron_cbor(cls: Type[Address], cbor_bytes: bytes) -> Address:
        """Decode a Byron address from CBOR bytes.

        Args:
            cbor_bytes: CBOR-encoded Byron address bytes.

        Returns:
            Address: Decoded Byron address instance.

        Raises:
            DecodingException: When decoding fails.
        """
        try:
            decoded = cbor2.loads(cbor_bytes)
        except Exception as e:
            raise DecodingException(f"Failed to decode CBOR bytes: {e}")

        # Byron address structure: [CBORTag(24, payload), crc32]
        if not isinstance(decoded, (tuple, list)) or len(decoded) != 2:
            raise DecodingException(
                f"Byron address must be a 2-element array, got {type(decoded)}"
            )

        tagged_payload, crc32_checksum = decoded

        if not isinstance(tagged_payload, CBORTag) or tagged_payload.tag != 24:
            raise DecodingException(
                f"Byron address must use CBOR tag 24, got {tagged_payload}"
            )

        payload_cbor = tagged_payload.value
        if not isinstance(payload_cbor, bytes):
            raise DecodingException(
                f"Tag 24 must contain bytes, got {type(payload_cbor)}"
            )

        computed_crc32 = binascii.crc32(payload_cbor) & 0xFFFFFFFF
        if computed_crc32 != crc32_checksum:
            raise DecodingException(
                f"CRC32 checksum mismatch: expected {crc32_checksum}, got {computed_crc32}"
            )

        try:
            payload = cbor2.loads(payload_cbor)
        except Exception as e:
            raise DecodingException(f"Failed to decode Byron address payload: {e}")

        if not isinstance(payload, (tuple, list)) or len(payload) != 3:
            raise DecodingException(
                f"Byron address payload must be a 3-element array, got {payload}"
            )

        payload_hash, attributes, byron_type = payload

        if not isinstance(payload_hash, bytes) or len(payload_hash) != 28:
            size = (
                len(payload_hash)
                if isinstance(payload_hash, bytes)
                else f"type {type(payload_hash).__name__}"
            )
            raise DecodingException(f"Payload hash must be 28 bytes, got {size}")

        if not isinstance(attributes, dict):
            raise DecodingException(
                f"Attributes must be a dict, got {type(attributes)}"
            )

        if byron_type not in (0, 2):
            raise DecodingException(f"Byron type must be 0 or 2, got {byron_type}")

        # Create Address instance with Byron fields
        addr = cls.__new__(cls)
        addr._payment_part = None
        addr._staking_part = None
        addr._byron_payload_hash = payload_hash
        addr._byron_attributes = attributes
        addr._byron_type = byron_type
        addr._byron_crc32 = crc32_checksum
        addr._network = addr._infer_byron_network()
        addr._address_type = AddressType.BYRON
        addr._header_byte = None
        addr._hrp = None
        return addr

    def _infer_byron_network(self) -> Network:
        """Infer network from Byron address attributes.

        Returns:
            Network: MAINNET or TESTNET (defaults to MAINNET).
        """
        if self._byron_attributes and 2 in self._byron_attributes:
            network_bytes = self._byron_attributes[2]
            if isinstance(network_bytes, bytes):
                try:
                    network_discriminant = cbor2.loads(network_bytes)
                    # Mainnet: 764824073 (0x2D964A09), Testnet: 1097911063 (0x42659F17)
                    if network_discriminant == 1097911063:
                        return Network.TESTNET
                except Exception:
                    pass
        return Network.MAINNET

    def __eq__(self, other):
        if not isinstance(other, Address):
            return False

        if self.is_byron != other.is_byron:
            return False

        if self.is_byron:
            return (
                self._byron_payload_hash == other._byron_payload_hash
                and self._byron_attributes == other._byron_attributes
                and self._byron_type == other._byron_type
                and self._byron_crc32 == other._byron_crc32
            )

        return (
            self.payment_part == other.payment_part
            and self.staking_part == other.staking_part
            and self.network == other.network
        )

    def __repr__(self):
        return f"{self.encode()}"

    @override
    def save(
        self,
        path: str,
        key_type: Optional[str] = None,
        description: Optional[str] = None,
        **kwargs,
    ):
        """
        Save the Address object to a file.

        This method writes the object's JSON representation to the specified file path.
         It raises an error if the file already exists and is not empty.

        Args:
            path (str): The file path to save the object to.
            key_type (str, optional): Not used in this context, but can be included for consistency.
            description (str, optional): Not used in this context, but can be included for consistency.
            **kwargs: Additional keyword arguments (not used here).

        Raises:
            IOError: If the file already exists and is not empty.
        """
        if os.path.isfile(path) and os.stat(path).st_size > 0:
            raise IOError(f"File {path} already exists!")
        with open(path, "w") as f:
            f.write(self.encode())

    @classmethod
    def load(cls, path: str) -> Address:
        """
        Load an Address object from a file.

        Args:
            path (str): The file path to load the object from.

        Returns:
            Address: The loaded Address object.
        """
        with open(path) as f:
            return cls.decode(f.read())



from __future__ import annotations

from dataclasses import dataclass, field
from enum import Enum, unique
from typing import Optional, Tuple, Type, Union

from pycardano.crypto.bech32 import bech32_decode, convertbits, encode
from pycardano.exception import (
    DecodingException,
    DeserializeException,
    SerializeException,
)
from pycardano.hash import (
    CIP129_PAYLOAD_SIZE,
    VERIFICATION_KEY_HASH_SIZE,
    AnchorDataHash,
    PoolKeyHash,
    ScriptHash,
    VerificationKeyHash,
)
from pycardano.serialization import (
    ArrayCBORSerializable,
    CodedSerializable,
    limit_primitive_type,
)

__all__ = [
    "Certificate",
    "StakeCredential",
    "StakeRegistration",
    "StakeDeregistration",
    "StakeDelegation",
    "PoolRegistration",
    "PoolRetirement",
    "StakeRegistrationConway",
    "StakeDeregistrationConway",
    "VoteDelegation",
    "StakeAndVoteDelegation",
    "StakeRegistrationAndDelegation",
    "StakeRegistrationAndVoteDelegation",
    "StakeRegistrationAndDelegationAndVoteDelegation",
    "DRep",
    "DRepKind",
    "AuthCommitteeHotCertificate",
    "ResignCommitteeColdCertificate",
    "Anchor",
    "DRepCredential",
    "RegDRepCert",
    "UnregDRepCertificate",
    "UpdateDRepCertificate",
    "GovernanceCredential",
    "GovernanceKeyType",
]

from pycardano.pool_params import PoolParams

unit_interval = Tuple[int, int]


@dataclass(repr=False)
class Anchor(ArrayCBORSerializable):
    """Represents an anchor in the Cardano blockchain that contains a URL and associated data hash.

    Anchors are used to provide additional metadata or reference external resources in certificates.
    """

    url: str
    """The URL pointing to the anchor's resource location"""

    data_hash: AnchorDataHash
    """The hash of the data associated with this anchor"""


@dataclass(repr=False)
class StakeCredential(ArrayCBORSerializable):
    """Represents a stake credential in the Cardano blockchain.

    A stake credential can either be a verification key hash or a script hash,
    used to identify stake rights and permissions.
    """

    _CODE: Optional[int] = field(init=False, default=None)

    credential: Union[VerificationKeyHash, ScriptHash]
    """The actual credential, either a verification key hash or script hash"""

    def __post_init__(self):
        if isinstance(self.credential, VerificationKeyHash):
            self._CODE = 0
        else:
            self._CODE = 1

    @classmethod
    @limit_primitive_type(list, tuple)
    def from_primitive(
        cls: Type[StakeCredential], values: Union[list, tuple]
    ) -> StakeCredential:
        if values[0] == 0:
            return cls(VerificationKeyHash(values[1]))
        elif values[0] == 1:
            return cls(ScriptHash(values[1]))
        else:
            raise DeserializeException(f"Invalid StakeCredential type {values[0]}")

    def __hash__(self):
        return hash(self.to_cbor())


class IdFormat(Enum):
    """
    Id format definition.
    """

    CIP129 = "cip129"
    CIP105 = "cip105"


class CredentialType(Enum):
    """
    Credential type definition.
    """

    KEY_HASH = 0b0010
    """Key hash"""

    SCRIPT_HASH = 0b0011
    """Script hash"""


class GovernanceKeyType(Enum):
    """
    Governance key type definition.
    """

    CC_HOT = 0b0000
    """Committee cold hot key"""

    CC_COLD = 0b0001
    """Committee cold key"""

    DREP = 0b0010
    """DRep key"""


@dataclass(repr=False)
class GovernanceCredential(StakeCredential):
    """Represents a governance credential."""

    governance_key_type: GovernanceKeyType = field(init=False)
    """Governance key type."""

    id_format: IdFormat = field(default=IdFormat.CIP129, compare=False)
    """Id format."""

    def __repr__(self):
        return f"{self.encode()}"

    def __bytes__(self):
        if self.id_format == IdFormat.CIP129:
            return self._compute_header_byte() + bytes(self.credential.payload)
        else:
            return bytes(self.credential.payload)

    @property
    def credential_type(self) -> CredentialType:
        """Credential type."""
        if isinstance(self.credential, VerificationKeyHash):
            return CredentialType.KEY_HASH
        else:
            return CredentialType.SCRIPT_HASH

    def _compute_header_byte(self) -> bytes:
        """Compute the header byte."""
        return (
            self.governance_key_type.value << 4 | self.credential_type.value
        ).to_bytes(1, byteorder="big")

    def _compute_hrp(self, id_format: IdFormat = IdFormat.CIP129) -> str:
        """Compute human-readable prefix for bech32 encoder.

        Based on
        `miscellaneous section <https://github.com/cardano-foundation/CIPs/tree/master/CIP-0005#miscellaneous>`_
        in CIP-5.
        """
        prefix = ""
        if self.governance_key_type == GovernanceKeyType.CC_HOT:
            prefix = "cc_hot"
        elif self.governance_key_type == GovernanceKeyType.CC_COLD:
            prefix = "cc_cold"
        elif self.governance_key_type == GovernanceKeyType.DREP:
            prefix = "drep"

        suffix = ""
        if isinstance(self.credential, VerificationKeyHash):
            suffix = ""
        elif isinstance(self.credential, ScriptHash):
            suffix = "_script"

        return prefix + suffix if id_format == IdFormat.CIP105 else prefix

    def encode(self) -> str:
        """Encode the governance credential in Bech32 format.

        More info about Bech32 `here <https://github.com/bitcoin/bips/blob/master/bip-0173.mediawiki#Bech32>`_.

        Returns:
            str: Encoded governance credential in Bech32 format.
        """
        data = bytes(self)
        return encode(self._compute_hrp(self.id_format), data)

    @classmethod
    def decode(cls: Type[GovernanceCredential], data: str) -> GovernanceCredential:
        """Decode a bech32 string into a governance credential object.

        Args:
            data (str): Bech32-encoded string.

        Returns:
            GovernanceCredential: Decoded governance credential.

        Raises:
            DecodingException: When the input string is not a valid governance credential.
        """
        hrp, checksum, _ = bech32_decode(data)
        value = bytes(convertbits(checksum, 5, 8, False))
        if len(value) == VERIFICATION_KEY_HASH_SIZE:
            # CIP-105
            if "script" in hrp:
                return cls(credential=ScriptHash(value))
            else:
                return cls(credential=VerificationKeyHash(value))
        elif len(value) == CIP129_PAYLOAD_SIZE:
            header = value[0]
            payload = value[1:]

            key_type = GovernanceKeyType((header & 0xF0) >> 4)
            credential_type = CredentialType(header & 0x0F)

            if key_type != cls.governance_key_type:
                raise DecodingException(f"Invalid key type: {key_type}")

            if credential_type == CredentialType.KEY_HASH:
                return cls(credential=VerificationKeyHash(payload))
            elif credential_type == CredentialType.SCRIPT_HASH:
                return cls(credential=ScriptHash(payload))
            else:
                raise DecodingException(f"Invalid credential type: {credential_type}")
        else:
            raise DecodingException(f"Invalid data length: {len(value)}")

    def to_primitive(self):
        return [self._CODE, self.credential.to_primitive()]


@dataclass(repr=False)
class DRepCredential(GovernanceCredential):
    """Represents a Delegate Representative (DRep) credential.

    This credential type is specifically used for DReps in the governance system,
    inheriting from GovernanceCredential.
    """

    governance_key_type: GovernanceKeyType = GovernanceKeyType.DREP


@unique
class DRepKind(Enum):
    """Enumerates the different types of Delegate Representatives (DReps).

    Defines the possible kinds of DReps in the Cardano governance system:
    - VERIFICATION_KEY_HASH: A DRep identified by a verification key hash
    - SCRIPT_HASH: A DRep identified by a script hash
    - ALWAYS_ABSTAIN: A special DRep that always abstains from voting
    - ALWAYS_NO_CONFIDENCE: A special DRep that always votes no confidence
    """

    VERIFICATION_KEY_HASH = 0
    SCRIPT_HASH = 1
    ALWAYS_ABSTAIN = 2
    ALWAYS_NO_CONFIDENCE = 3


@dataclass(repr=False)
class DRep(ArrayCBORSerializable):
    """Represents a Delegate Representative (DRep) in the Cardano governance system.

    DReps are entities that can represent stake holders in governance decisions.
    """

    kind: DRepKind
    """The type of DRep (verification key, script hash, always abstain, or always no confidence)"""

    credential: Optional[Union[VerificationKeyHash, ScriptHash]] = field(
        default=None, metadata={"optional": True}
    )
    """The credential associated with this DRep, if applicable"""

    id_format: IdFormat = field(default=IdFormat.CIP129, compare=False)

    def __repr__(self):
        return f"{self.encode()}"

    def __bytes__(self):
        if self.credential is not None:
            drep_credential = DRepCredential(
                credential=self.credential, id_format=self.id_format
            )
            return bytes(drep_credential)
        return b""

    @classmethod
    @limit_primitive_type(list)
    def from_primitive(cls: Type[DRep], values: Union[list, tuple]) -> DRep:
        try:
            kind = DRepKind(values[0])
        except ValueError as e:
            raise DeserializeException(f"Invalid DRep type {values[0]}") from e

        if kind == DRepKind.VERIFICATION_KEY_HASH:
            return cls(kind=kind, credential=VerificationKeyHash(values[1]))
        elif kind == DRepKind.SCRIPT_HASH:
            return cls(kind=kind, credential=ScriptHash(values[1]))
        elif kind == DRepKind.ALWAYS_ABSTAIN:
            return cls(kind=kind)
        elif kind == DRepKind.ALWAYS_NO_CONFIDENCE:
            return cls(kind=kind)
        else:
            raise DeserializeException(f"Invalid DRep type {values[0]}")

    def to_primitive(self):
        if self.credential is not None:
            return [self.kind.value, self.credential.to_primitive()]
        return [self.kind.value]

    def encode(self) -> str:
        """Encode the DRep in Bech32 format.

        More info about Bech32 `here <https://github.com/bitcoin/bips/blob/master/bip-0173.mediawiki#Bech32>`_.

        Returns:
            str: Encoded DRep in Bech32 format.

        Examples:
            >>> vkey_bytes = bytes.fromhex("00000000000000000000000000000000000000000000000000000000")
            >>> credential = VerificationKeyHash(vkey_bytes)
            >>> print(DRep(kind=DRepKind.VERIFICATION_KEY_HASH, credential=credential).encode())
            drep1ygqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqq7vlc9n
        """
        if self.kind == DRepKind.ALWAYS_ABSTAIN:
            return "drep_always_abstain"
        elif self.kind == DRepKind.ALWAYS_NO_CONFIDENCE:
            return "drep_always_no_confidence"
        elif self.credential is not None:
            drep_credential = DRepCredential(
                credential=self.credential, id_format=self.id_format
            )
            return drep_credential.encode()
        else:
            raise SerializeException("DRep credential is None")

    @classmethod
    def decode(cls: Type[DRep], data: str) -> DRep:
        """Decode a bech32 string into a DRep object.

        Args:
            data (str): Bech32-encoded string.

        Returns:
            DRep: Decoded DRep.

        Raises:
            DecodingException: When the input string is not a valid DRep.

        Examples:
            >>> credential = DRep.decode("drep1ygqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqq7vlc9n")
            >>> khash = VerificationKeyHash(bytes.fromhex("00000000000000000000000000000000000000000000000000000000"))
            >>> assert credential == DRep(DRepKind.VERIFICATION_KEY_HASH, khash)
        """
        if data == "drep_always_abstain":
            return cls(kind=DRepKind.ALWAYS_ABSTAIN)
        elif data == "drep_always_no_confidence":
            return cls(kind=DRepKind.ALWAYS_NO_CONFIDENCE)
        else:
            drep_credential = DRepCredential.decode(data)
            return cls(
                kind=(
                    DRepKind.VERIFICATION_KEY_HASH
                    if isinstance(drep_credential.credential, VerificationKeyHash)
                    else DRepKind.SCRIPT_HASH
                ),
                credential=drep_credential.credential,
            )


@dataclass(repr=False)
class StakeRegistration(CodedSerializable):
    """Certificate for registering a stake credential."""

    _CODE: int = field(init=False, default=0)
    stake_credential: StakeCredential
    """The stake credential being registered"""


@dataclass(repr=False)
class StakeDeregistration(CodedSerializable):
    """Certificate for deregistering a stake credential."""

    _CODE: int = field(init=False, default=1)
    stake_credential: StakeCredential
    """The stake credential being deregistered"""


@dataclass(repr=False)
class StakeDelegation(CodedSerializable):
    """Certificate for delegating stake to a stake pool."""

    _CODE: int = field(init=False, default=2)
    stake_credential: StakeCredential
    """The stake credential being delegated"""

    pool_keyhash: PoolKeyHash
    """The hash of the pool's key to delegate to"""


@dataclass(repr=False)
class PoolRegistration(CodedSerializable):
    """Certificate for registering a stake pool."""

    _CODE: int = field(init=False, default=3)
    pool_params: PoolParams
    """The parameters defining the stake pool's configuration"""

    def to_primitive(self):
        pool_params = self.pool_params.to_primitive()
        if isinstance(pool_params, list):
            return [self._CODE, *pool_params]
        return super().to_primitive()

    @classmethod
    @limit_primitive_type(list, tuple)
    def from_primitive(
        cls: Type[PoolRegistration], values: Union[list, tuple]
    ) -> PoolRegistration:
        if values[0] == 3:
            if isinstance(values[1], list):
                return cls(
                    pool_params=PoolParams.from_primitive(values[1]),
                )
            else:
                return cls(
                    pool_params=PoolParams.from_primitive(values[1:]),
                )
        else:
            raise DeserializeException(f"Invalid PoolRegistration type {values[0]}")


@dataclass(repr=False)
class PoolRetirement(CodedSerializable):
    """Certificate for retiring a stake pool."""

    _CODE: int = field(init=False, default=4)
    pool_keyhash: PoolKeyHash
    """The hash of the pool's key that is being retired"""

    epoch: int
    """The epoch number when the pool will retire"""


@dataclass(repr=False)
class StakeRegistrationConway(CodedSerializable):
    """Certificate for registering a stake credential in the Conway era."""

    _CODE: int = field(init=False, default=7)
    stake_credential: StakeCredential
    """The stake credential being registered"""

    coin: int
    """The amount of coins associated with this registration"""


@dataclass(repr=False)
class StakeDeregistrationConway(CodedSerializable):
    """Certificate for deregistering a stake credential in the Conway era."""

    _CODE: int = field(init=False, default=8)
    stake_credential: StakeCredential
    """The stake credential being deregistered"""

    coin: int
    """The amount of coins associated with this deregistration"""


@dataclass(repr=False)
class VoteDelegation(CodedSerializable):
    """Certificate for delegating voting power to a DRep."""

    _CODE: int = field(init=False, default=9)
    stake_credential: StakeCredential
    """The stake credential delegating its voting power"""

    drep: DRep
    """The DRep receiving the voting power delegation"""


@dataclass(repr=False)
class StakeAndVoteDelegation(CodedSerializable):
    """Certificate for delegating both stake and voting power."""

    _CODE: int = field(init=False, default=10)
    stake_credential: StakeCredential
    """The stake credential being delegated"""

    pool_keyhash: PoolKeyHash
    """The hash of the pool's key receiving the stake delegation"""

    drep: DRep
    """The DRep receiving the voting power delegation"""


@dataclass(repr=False)
class StakeRegistrationAndDelegation(CodedSerializable):
    """Certificate for registering stake and delegating to a pool."""

    _CODE: int = field(init=False, default=11)
    stake_credential: StakeCredential
    """The stake credential being registered and delegated"""

    pool_keyhash: PoolKeyHash
    """The hash of the pool's key receiving the delegation"""

    coin: int
    """The amount of coins associated with this registration"""


@dataclass(repr=False)
class StakeRegistrationAndVoteDelegation(CodedSerializable):
    """Certificate for registering stake and delegating voting power."""

    _CODE: int = field(init=False, default=12)
    stake_credential: StakeCredential
    """The stake credential being registered"""

    drep: DRep
    """The DRep receiving the voting power delegation"""

    coin: int
    """The amount of coins associated with this registration"""


@dataclass(repr=False)
class StakeRegistrationAndDelegationAndVoteDelegation(CodedSerializable):
    """Certificate for registering stake and delegating both stake and voting power."""

    _CODE: int = field(init=False, default=13)
    stake_credential: StakeCredential
    """The stake credential being registered and delegated"""

    pool_keyhash: PoolKeyHash
    """The hash of the pool's key receiving the stake delegation"""

    drep: DRep
    """The DRep receiving the voting power delegation"""

    coin: int
    """The amount of coins associated with this registration"""


@dataclass(repr=False)
class AuthCommitteeHotCertificate(CodedSerializable):
    """Certificate for authorizing a committee hot key."""

    _CODE: int = field(init=False, default=14)
    committee_cold_credential: StakeCredential
    """The cold credential of the committee member"""

    committee_hot_credential: StakeCredential
    """The hot credential being authorized"""


@dataclass(repr=False)
class ResignCommitteeColdCertificate(CodedSerializable):
    """Certificate for resigning from the constitutional committee."""

    _CODE: int = field(init=False, default=15)
    committee_cold_credential: StakeCredential
    """The cold credential of the resigning committee member"""

    anchor: Optional[Anchor]
    """Optional anchor containing additional metadata about the resignation"""


@dataclass(repr=False)
class RegDRepCert(CodedSerializable):
    """Certificate for registering as a delegate representative (DRep)."""

    _CODE: int = field(init=False, default=16)
    drep_credential: DRepCredential
    """The credential of the DRep being registered"""

    coin: int
    """The amount of coins associated with this registration"""

    anchor: Optional[Anchor] = field(default=None)
    """Optional anchor containing additional metadata about the DRep"""


@dataclass(repr=False)
class UnregDRepCertificate(CodedSerializable):
    """Certificate for unregistering as a delegate representative (DRep)."""

    _CODE: int = field(init=False, default=17)
    drep_credential: DRepCredential
    """The credential of the DRep being unregistered"""

    coin: int
    """The amount of coins associated with this unregistration"""


@dataclass(repr=False)
class UpdateDRepCertificate(CodedSerializable):
    """Certificate for updating delegate representative (DRep) metadata."""

    _CODE: int = field(init=False, default=18)
    drep_credential: DRepCredential
    """The credential of the DRep being updated"""

    anchor: Optional[Anchor]
    """Optional anchor containing the updated metadata"""


Certificate = Union[
    StakeRegistration,
    StakeDeregistration,
    StakeDelegation,
    PoolRegistration,
    PoolRetirement,
    StakeRegistrationConway,
    StakeDeregistrationConway,
    VoteDelegation,
    StakeAndVoteDelegation,
    StakeRegistrationAndDelegation,
    StakeRegistrationAndVoteDelegation,
    StakeRegistrationAndDelegationAndVoteDelegation,
    AuthCommitteeHotCertificate,
    ResignCommitteeColdCertificate,
    UnregDRepCertificate,
    UpdateDRepCertificate,
]



from __future__ import annotations

from dataclasses import dataclass, field
from enum import Enum, unique
from fractions import Fraction
from typing import Dict, Optional, Tuple, Type, Union

from pycardano.certificate import Anchor, GovernanceCredential, GovernanceKeyType
from pycardano.crypto.bech32 import bech32_decode, convertbits, encode
from pycardano.exception import (
    DecodingException,
    DeserializeException,
    InvalidDataException,
)
from pycardano.hash import PolicyHash, ScriptHash, TransactionId, VerificationKeyHash
from pycardano.plutus import ExecutionUnits
from pycardano.serialization import (
    ArrayCBORSerializable,
    CodedSerializable,
    DictCBORSerializable,
    MapCBORSerializable,
    OrderedSet,
    Primitive,
    limit_primitive_type,
)

__all__ = [
    "CommitteeColdCredential",
    "CommitteeHotCredential",
    "CommitteeColdCredentialEpochMap",
    "ParameterChangeAction",
    "HardForkInitiationAction",
    "TreasuryWithdrawalsAction",
    "NoConfidence",
    "UpdateCommittee",
    "NewConstitution",
    "InfoAction",
    "GovActionId",
    "VotingProcedure",
    "VotingProcedures",
    "VoterType",
    "Voter",
    "GovAction",
    "Vote",
    "Anchor",
    "GovActionIdToVotingProcedure",
    "ProposalProcedure",
    "ProtocolParamUpdate",
    "ExUnitPrices",
    "DRepVotingThresholds",
    "TreasuryWithdrawal",
    "PoolVotingThresholds",
]


class CommitteeColdCredential(GovernanceCredential):
    """Represents a cold credential for a committee member."""

    governance_key_type: GovernanceKeyType = GovernanceKeyType.CC_COLD


class CommitteeHotCredential(GovernanceCredential):
    """Represents a hot credential for a committee member."""

    governance_key_type: GovernanceKeyType = GovernanceKeyType.CC_HOT


@unique
class Vote(Enum):
    """Represents possible voting choices in the governance system."""

    NO = 0
    YES = 1
    ABSTAIN = 2


@dataclass(repr=False)
class GovActionId(ArrayCBORSerializable):
    """Represents a unique identifier for a governance action.

    This identifier consists of a transaction ID and an index within that transaction.
    """

    transaction_id: TransactionId
    """The transaction ID where this governance action was submitted"""

    gov_action_index: int
    """The index of this governance action within the transaction (0-65535)"""

    def __post_init__(self):
        if not 0 <= self.gov_action_index <= 65535:  # uint .size 2
            raise ValueError("gov_action_index must be between 0 and 65535")

    @classmethod
    @limit_primitive_type(list, tuple)
    def from_primitive(
        cls: Type[GovActionId], values: Union[list, tuple]
    ) -> GovActionId:
        return cls(transaction_id=TransactionId(values[0]), gov_action_index=values[1])

    def __hash__(self):
        return hash((self.transaction_id, self.gov_action_index))

    def __repr__(self):
        return f"{self.encode()}"

    def __bytes__(self):
        # Convert index to hex (no prefix, lowercase)
        idx_hex = f"{self.gov_action_index:x}"

        # Pad to even-length hex
        if len(idx_hex) % 2 != 0:
            idx_hex = f"0{idx_hex}"

        try:
            idx_bytes = bytes.fromhex(idx_hex)
            return self.transaction_id.payload + idx_bytes
        except ValueError as e:
            raise InvalidDataException(f"Error encoding data: {idx_hex}") from e

    def encode(self) -> str:
        """Encode the governance action ID in Bech32 format.

        More info about Bech32 `here <https://github.com/bitcoin/bips/blob/master/bip-0173.mediawiki#Bech32>`_.

        Returns:
            str: Encoded pool key hash in Bech32.

        Examples:
            >>> transaction_id = TransactionId(bytes.fromhex("00" * 32))
            >>> gov_action_id = GovActionId(transaction_id=transaction_id, gov_action_index=17)
            >>> print(gov_action_id.encode())
            gov_action1qqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqpzklpgpf
        """
        return encode(
            "gov_action",
            bytes(self),
        )

    @classmethod
    def decode(cls, data: str) -> GovActionId:
        """Decode a bech32 string into a governance action ID object.

        Args:
            data (str): Bech32-encoded string.

        Returns:
            GovActionId: Decoded governance action ID.

        Raises:
            DecodingException: When the input string is not a valid governance action ID.

        Examples:
            >>> bech32_id = "gov_action1qqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqpzklpgpf"
            >>> gov_action_id = GovActionId.decode(bech32_id)
            >>> transaction_id = TransactionId(bytes.fromhex("00" * 32))
            >>> assert gov_action_id == GovActionId(transaction_id, 17)
        """
        hrp, checksum, _ = bech32_decode(data)
        value = bytes(convertbits(checksum, 5, 8, False))

        if hrp != "gov_action":
            raise DecodingException("Invalid GovActionId bech32 string")

        # Transaction ID is always 32 bytes, the rest is the index
        if len(value) < 33:
            raise DecodingException(
                f"Invalid GovActionId length: {len(value)}, expected at least 33 bytes"
            )

        tx_id = TransactionId(value[:32])
        index_bytes = value[32:]
        index = int.from_bytes(index_bytes, "big")
        return cls(transaction_id=tx_id, gov_action_index=index)


@dataclass(repr=False)
class ExUnitPrices(ArrayCBORSerializable):
    """Represents execution unit prices for memory and CPU steps."""

    mem_price: Fraction
    """Memory price as a nonnegative interval (numerator, denominator)"""

    step_price: Fraction
    """Step price as a nonnegative interval (numerator, denominator)"""


@dataclass(repr=False)
class DRepVotingThresholds(ArrayCBORSerializable):
    """Represents DRep voting thresholds as an array of unit intervals."""

    motion_no_confidence: Fraction
    """Threshold for no confidence motions"""

    committee_normal: Fraction
    """Threshold for normal committee updates"""

    committee_no_confidence: Fraction
    """Threshold for committee updates during no confidence"""

    update_constitution: Fraction
    """Threshold for constitution updates"""

    hard_fork_initiation: Fraction
    """Threshold for hard fork initiation"""

    pp_network_group: Fraction
    """Threshold for network group protocol parameter updates"""

    pp_economic_group: Fraction
    """Threshold for economic group protocol parameter updates"""

    pp_technical_group: Fraction
    """Threshold for technical group protocol parameter updates"""

    pp_governance_group: Fraction
    """Threshold for governance group protocol parameter updates"""

    treasury_withdrawal: Fraction
    """Threshold for treasury withdrawals"""


@dataclass(repr=False)
class PoolVotingThresholds(ArrayCBORSerializable):
    """Represents pool voting thresholds as an array of unit intervals."""

    motion_no_confidence: Fraction
    """Threshold for no confidence motions"""

    committee_normal: Fraction
    """Threshold for normal committee updates"""

    committee_no_confidence: Fraction
    """Threshold for committee updates during no confidence"""

    hard_fork_initiation: Fraction
    """Threshold for hard fork initiation"""

    ppec_voting_threshold: Fraction
    """Threshold for protocol parameter and economic group voting"""


@dataclass(repr=False)
class ProtocolParamUpdate(MapCBORSerializable):
    """Represents a protocol parameter update."""

    min_fee_a: Optional[int] = field(
        default=None, metadata={"key": 0, "optional": True}
    )
    """The 'a' parameter to calculate the minimum fee"""

    min_fee_b: Optional[int] = field(
        default=None, metadata={"key": 1, "optional": True}
    )
    """The 'b' parameter to calculate the minimum fee"""

    max_block_body_size: Optional[int] = field(
        default=None, metadata={"key": 2, "optional": True}
    )
    """Maximum block body size"""

    max_transaction_size: Optional[int] = field(
        default=None, metadata={"key": 3, "optional": True}
    )
    """Maximum transaction size"""

    max_block_header_size: Optional[int] = field(
        default=None, metadata={"key": 4, "optional": True}
    )
    """Maximum block header size"""

    key_deposit: Optional[int] = field(
        default=None, metadata={"key": 5, "optional": True}
    )
    """The deposit required for registering a stake key"""

    pool_deposit: Optional[int] = field(
        default=None, metadata={"key": 6, "optional": True}
    )
    """The deposit required for registering a stake pool"""

    maximum_epoch: Optional[int] = field(
        default=None, metadata={"key": 7, "optional": True}
    )
    """Maximum epoch interval (uint32)"""

    n_opt: Optional[int] = field(default=None, metadata={"key": 8, "optional": True})
    """Desired number of stake pools"""

    pool_pledge_influence: Optional[Fraction] = field(
        default=None, metadata={"key": 9, "optional": True}
    )
    """Pool pledge influence as a nonnegative interval (numerator, denominator)"""

    expansion_rate: Optional[Fraction] = field(
        default=None, metadata={"key": 10, "optional": True}
    )
    """Monetary expansion rate as a unit interval (numerator, denominator)"""

    treasury_growth_rate: Optional[Fraction] = field(
        default=None, metadata={"key": 11, "optional": True}
    )
    """Treasury growth rate as a unit interval (numerator, denominator)"""

    min_pool_cost: Optional[int] = field(
        default=None, metadata={"key": 16, "optional": True}
    )
    """Minimum pool cost"""

    ada_per_utxo_byte: Optional[int] = field(
        default=None, metadata={"key": 17, "optional": True}
    )
    """Ada per UTxO byte"""

    cost_models: Optional[Dict] = field(
        default=None, metadata={"key": 18, "optional": True}
    )
    """Cost models for script languages"""

    execution_costs: Optional[ExUnitPrices] = field(
        default=None, metadata={"key": 19, "optional": True}
    )
    """Execution costs (prices) for ex units"""

    max_tx_ex_units: Optional[ExecutionUnits] = field(
        default=None, metadata={"key": 20, "optional": True}
    )
    """Maximum execution units per transaction"""

    max_block_ex_units: Optional[ExecutionUnits] = field(
        default=None, metadata={"key": 21, "optional": True}
    )
    """Maximum execution units per block"""

    max_value_size: Optional[int] = field(
        default=None, metadata={"key": 22, "optional": True}
    )
    """Maximum value size"""

    collateral_percentage: Optional[int] = field(
        default=None, metadata={"key": 23, "optional": True}
    )
    """Collateral percentage"""

    max_collateral_inputs: Optional[int] = field(
        default=None, metadata={"key": 24, "optional": True}
    )
    """Maximum number of collateral inputs"""

    pool_voting_thresholds: Optional[PoolVotingThresholds] = field(
        default=None, metadata={"key": 25, "optional": True}
    )
    """Pool voting thresholds"""

    drep_voting_thresholds: Optional[DRepVotingThresholds] = field(
        default=None, metadata={"key": 26, "optional": True}
    )
    """DRep voting thresholds"""

    min_committee_size: Optional[int] = field(
        default=None, metadata={"key": 27, "optional": True}
    )
    """Minimum committee size"""

    committee_term_limit: Optional[int] = field(
        default=None, metadata={"key": 28, "optional": True}
    )
    """Committee term limit in epochs (uint32)"""

    governance_action_validity_period: Optional[int] = field(
        default=None, metadata={"key": 29, "optional": True}
    )
    """Governance action validity period in epochs (uint32)"""

    governance_action_deposit: Optional[int] = field(
        default=None, metadata={"key": 30, "optional": True}
    )
    """Deposit required for governance actions"""

    drep_deposit: Optional[int] = field(
        default=None, metadata={"key": 31, "optional": True}
    )
    """Deposit required for DRep registration"""

    drep_inactivity_period: Optional[int] = field(
        default=None, metadata={"key": 32, "optional": True}
    )
    """DRep inactivity period in epochs (uint32)"""

    min_fee_ref_script_cost: Optional[Fraction] = field(
        default=None, metadata={"key": 33, "optional": True}
    )
    """Minimum fee for reference scripts as a nonnegative interval (numerator, denominator)"""


@dataclass(repr=False)
class ParameterChangeAction(CodedSerializable):
    """Represents a governance action to change protocol parameters."""

    _CODE: int = field(init=False, default=0)
    gov_action_id: Optional[GovActionId]
    """Optional reference to a previous governance action"""

    protocol_param_update: ProtocolParamUpdate
    """Dictionary containing the protocol parameters to be updated"""

    policy_hash: Optional[PolicyHash]
    """Optional policy hash associated with this parameter change"""


@dataclass(repr=False)
class HardForkInitiationAction(CodedSerializable):
    """Represents a governance action to initiate a hard fork."""

    _CODE: int = field(init=False, default=1)
    gov_action_id: Optional[GovActionId]
    """Optional reference to a previous governance action"""

    protocol_version: Fraction
    """The target protocol version as (major, minor) version numbers"""

    def __post_init__(self):
        major, minor = self.protocol_version
        if not 1 <= major <= 10:
            raise ValueError("Major protocol version must be between 1 and 10")


class TreasuryWithdrawal(DictCBORSerializable):
    """Represents a treasury withdrawal amount and destination."""

    KEY_TYPE = bytes

    VALUE_TYPE = int


@dataclass(repr=False)
class TreasuryWithdrawalsAction(CodedSerializable):
    """Represents a governance action to withdraw funds from the treasury."""

    _CODE: int = field(init=False, default=2)
    withdrawals: TreasuryWithdrawal
    """The withdrawal amounts and their destinations"""

    policy_hash: Optional[PolicyHash]
    """Optional policy hash associated with these withdrawals"""


@dataclass(repr=False)
class NoConfidence(CodedSerializable):
    """Represents a governance action expressing no confidence."""

    _CODE: int = field(init=False, default=3)
    gov_action_id: Optional[GovActionId]
    """Optional reference to a previous governance action"""


class CommitteeColdCredentialEpochMap(DictCBORSerializable):
    """Represents a mapping of committee members to their expiration epochs."""

    KEY_TYPE = CommitteeColdCredential
    VALUE_TYPE = int


@dataclass(repr=False)
class UpdateCommittee(CodedSerializable):
    """Represents a governance action to update the constitutional committee."""

    _CODE: int = field(init=False, default=4)
    gov_action_id: Optional[GovActionId]
    """Optional reference to a previous governance action"""

    committee_cold_credentials: OrderedSet[CommitteeColdCredential]
    """Set of cold credentials for committee members"""

    committee_expiration: CommitteeColdCredentialEpochMap
    """Mapping of committee members to their expiration epochs"""

    quorum: Fraction  # unit_interval
    """The quorum threshold as a unit interval (numerator, denominator)"""


@dataclass(repr=False)
class NewConstitution(CodedSerializable):
    """Represents a governance action to establish a new constitution."""

    _CODE: int = field(init=False, default=5)
    gov_action_id: Optional[GovActionId]
    """Optional reference to a previous governance action"""

    constitution: Tuple[Anchor, Optional[ScriptHash]]
    """The constitution data as (anchor, optional script hash)"""


@dataclass(repr=False)
class InfoAction(CodedSerializable):
    """Represents an informational governance action with no direct effect."""

    _CODE: int = field(init=False, default=6)


@dataclass(repr=False)
class VotingProcedure(ArrayCBORSerializable):
    """Represents a voting procedure for governance actions.

    This defines how voting is conducted for a specific governance action.
    """

    vote: Vote
    """The vote cast (YES, NO, or ABSTAIN)"""

    anchor: Optional[Anchor]
    """Optional metadata associated with this vote"""

    @classmethod
    @limit_primitive_type(list)
    def from_primitive(
        cls: Type[VotingProcedure], values: Union[list, tuple]
    ) -> VotingProcedure:
        return cls(
            vote=Vote(values[0]),
            anchor=Anchor.from_primitive(values[1]) if values[1] is not None else None,
        )

    def to_shallow_primitive(self) -> Primitive:
        return [self.vote.value, self.anchor]


@unique
class VoterType(Enum):
    """Represents the possible types of voters in the governance system."""

    COMMITTEE_HOT = "committee_hot"
    DREP = "drep"
    STAKING_POOL = "staking_pool"


@dataclass(repr=False)
class Voter(ArrayCBORSerializable):
    """Represents a voter in the governance system.

    Voters can be committee members, DReps, or stake pool operators.
    """

    _CODE: Optional[int] = field(init=False, default=None)

    credential: Union[VerificationKeyHash, ScriptHash]
    """The credential identifying the voter"""

    voter_type: VoterType
    """The type of voter (COMMITTEE_HOT, DREP, or STAKING_POOL)"""

    def __post_init__(self):
        if self.voter_type == VoterType.COMMITTEE_HOT:
            if isinstance(self.credential, VerificationKeyHash):
                self._CODE = 0
            else:
                self._CODE = 1
        elif self.voter_type == VoterType.DREP:
            if isinstance(self.credential, VerificationKeyHash):
                self._CODE = 2
            else:
                self._CODE = 3
        elif self.voter_type == VoterType.STAKING_POOL:
            if not isinstance(self.credential, VerificationKeyHash):
                raise ValueError("Staking pool voter must use key hash credential")
            self._CODE = 4
        else:
            raise ValueError("Invalid voter_type")

    def to_shallow_primitive(self) -> Primitive:
        return self._CODE, self.credential

    @classmethod
    @limit_primitive_type(list, tuple)
    def from_primitive(cls: Type[Voter], values: Union[list, tuple]) -> Voter:
        code = values[0]
        credential: Union[VerificationKeyHash, ScriptHash]
        if code in (0, 2, 4):
            credential = VerificationKeyHash(values[1])
        elif code in (1, 3):
            credential = ScriptHash(values[1])
        else:
            raise DeserializeException(f"Invalid Voter type {code}")

        voter_type = {
            0: VoterType.COMMITTEE_HOT,
            1: VoterType.COMMITTEE_HOT,
            2: VoterType.DREP,
            3: VoterType.DREP,
            4: VoterType.STAKING_POOL,
        }[code]

        return cls(credential=credential, voter_type=voter_type)

    def __hash__(self):
        return hash((self._CODE, self.credential))


class GovActionIdToVotingProcedure(DictCBORSerializable):
    """Represents a mapping of governance action IDs to their voting procedures."""

    KEY_TYPE = GovActionId
    VALUE_TYPE = VotingProcedure


class VotingProcedures(DictCBORSerializable):
    """Represents a mapping of voters to their voting procedures."""

    KEY_TYPE = Voter
    VALUE_TYPE = GovActionIdToVotingProcedure


GovAction = Union[
    ParameterChangeAction,
    HardForkInitiationAction,
    TreasuryWithdrawalsAction,
    NoConfidence,
    UpdateCommittee,
    NewConstitution,
    InfoAction,
]


@dataclass(repr=False)
class ProposalProcedure(ArrayCBORSerializable):
    """Represents a proposal procedure for governance actions."""

    deposit: int
    """The deposit required to submit a proposal"""

    reward_account: bytes
    """The reward account for the proposal"""

    gov_action: GovAction
    """The governance actions to be proposed"""

    anchor: Anchor
    """The metadata anchor for the proposal"""



"""Defines CBOR serialization interfaces and provides useful serialization classes."""

from __future__ import annotations

import json
import os
import re
import typing
from collections import OrderedDict, UserList, defaultdict
from copy import deepcopy
from dataclasses import Field, dataclass, field, fields
from datetime import datetime
from decimal import Decimal
from fractions import Fraction
from functools import wraps
from inspect import getfullargspec, isclass
from typing import (
    Any,
    Callable,
    ClassVar,
    Dict,
    Generic,
    Iterable,
    List,
    Optional,
    Sequence,
    Type,
    TypeVar,
    Union,
    cast,
    get_type_hints,
)

from pycardano.cbor import cbor2
from pycardano.logging import logger

# Remove the semantic decoder for 258 (CBOR tag for set) as we care about the order of elements
try:
    cbor2._decoder.semantic_decoders.pop(258)
except Exception as e:
    logger.warning("Failed to remove semantic decoder for CBOR tag 258", e)
    pass

from cbor2 import CBOREncoder, CBORSimpleValue, CBORTag, FrozenDict, dumps, undefined
from frozenlist import FrozenList
from pprintpp import pformat

from pycardano.exception import DeserializeException, SerializeException
from pycardano.types import check_type, typechecked

__all__ = [
    "default_encoder",
    "IndefiniteList",
    "IndefiniteFrozenList",
    "Primitive",
    "CBORBase",
    "CBORSerializable",
    "ArrayCBORSerializable",
    "MapCBORSerializable",
    "DictCBORSerializable",
    "RawCBOR",
    "list_hook",
    "limit_primitive_type",
    "OrderedSet",
    "NonEmptyOrderedSet",
    "CodedSerializable",
]

T = TypeVar("T")


def _identity(x):
    return x


class IndefiniteList(UserList):
    def __init__(self, li: Primitive):  # type: ignore
        super().__init__(li)  # type: ignore


class IndefiniteFrozenList(FrozenList, IndefiniteList):  # type: ignore
    pass


@dataclass
class ByteString:
    value: bytes

    def __hash__(self):
        return hash(self.value)

    def __eq__(self, other: object):
        if isinstance(other, ByteString):
            return self.value == other.value
        elif isinstance(other, bytes):
            return self.value == other
        else:
            return False


@dataclass
class RawCBOR:
    """A wrapper class for bytes that represents a CBOR value."""

    cbor: bytes


Primitive = Union[
    bytes,
    bytearray,
    str,
    int,
    float,
    Decimal,
    bool,
    None,
    tuple,
    list,
    IndefiniteList,
    dict,
    defaultdict,
    OrderedDict,
    datetime,
    re.Pattern,
    CBORSimpleValue,
    CBORTag,
    set,
    Fraction,
    frozenset,
    FrozenDict,
    FrozenList,
    IndefiniteFrozenList,
    ByteString,
]

PRIMITIVE_TYPES = (
    bytes,
    bytearray,
    str,
    int,
    float,
    Decimal,
    bool,
    type(None),
    tuple,
    list,
    IndefiniteList,
    dict,
    defaultdict,
    OrderedDict,
    type(undefined),
    datetime,
    re.Pattern,
    CBORSimpleValue,
    CBORTag,
    set,
    frozenset,
    FrozenDict,
    Fraction,
    FrozenList,
    IndefiniteFrozenList,
    ByteString,
)
"""
A list of types that could be encoded by
`Cbor2 encoder <https://cbor2.readthedocs.io/en/latest/modules/encoder.html>`_ directly.
"""


def limit_primitive_type(*allowed_types):
    """
    A helper function to validate primitive type given to from_primitive class methods

    Not exposed to public by intention.
    """

    def decorator(func):
        @wraps(func)
        def wrapper(cls, value: Primitive):
            if not isinstance(value, allowed_types):
                allowed_types_str = [
                    allowed_type.__name__ for allowed_type in allowed_types
                ]
                raise DeserializeException(
                    f"{allowed_types_str} typed value is required for deserialization. Got {type(value)}: {value}"
                )
            return func(cls, value)

        return wrapper

    return decorator


CBORBase = TypeVar("CBORBase", bound="CBORSerializable")


def decode_array(self, subtype: int) -> Sequence[Any]:
    # Major tag 4
    length = self._decode_length(subtype, allow_indefinite=True)

    if length is None:
        ret = IndefiniteFrozenList(list(self.decode_array(subtype=subtype)))
        ret.freeze()
        return ret
    else:
        return self.decode_array(subtype=subtype)


try:
    cbor2._decoder.major_decoders[4] = decode_array
except Exception as e:
    logger.warning("Failed to replace major decoder for indefinite array", e)


def default_encoder(
    encoder: CBOREncoder, value: Union[CBORSerializable, IndefiniteList]
):
    """A fallback function that encodes CBORSerializable to CBOR"""
    assert isinstance(
        value,
        (
            ByteString,
            CBORSerializable,
            IndefiniteList,
            RawCBOR,
            FrozenList,
            IndefiniteFrozenList,
            FrozenDict,
        ),
    ), (
        f"Type of input value is not CBORSerializable, " f"got {type(value)} instead."
    )
    if isinstance(value, (IndefiniteList, IndefiniteFrozenList)):
        # Currently, cbor2 doesn't support indefinite list, therefore we need special
        # handling here to explicitly write header (b'\x9f'), each body item, and footer (b'\xff') to
        # the output bytestring.
        encoder.write(b"\x9f")
        for item in value:
            encoder.encode(item)
        encoder.write(b"\xff")
    elif isinstance(value, ByteString):
        if len(value.value) > 64:
            encoder.write(b"\x5f")
            for i in range(0, len(value.value), 64):
                imax = min(i + 64, len(value.value))
                encoder.encode(value.value[i:imax])
            encoder.write(b"\xff")
        else:
            encoder.encode(value.value)
    elif isinstance(value, RawCBOR):
        encoder.write(value.cbor)
    elif isinstance(value, FrozenList):
        encoder.encode(list(value))
    elif isinstance(value, FrozenDict):
        encoder.encode(dict(value))
    else:
        encoder.encode(value.to_validated_primitive())


@typechecked
class CBORSerializable:
    """
    CBORSerializable standardizes the interfaces a class should implement in order for it to be serialized to and
    deserialized from CBOR.

    Two required interfaces to implement are :meth:`to_primitive` and :meth:`from_primitive`.
    :meth:`to_primitive` converts an object to a CBOR primitive type (see :const:`Primitive`), which could be then
    encoded by CBOR library. :meth:`from_primitive` restores an object from a CBOR primitive type.

    To convert a CBORSerializable to CBOR, use :meth:`to_cbor`.
    To restore a CBORSerializable from CBOR, use :meth:`from_cbor`.

    .. note::
        :meth:`to_primitive` needs to return a pure CBOR primitive type, meaning that the returned value and all its
        child elements have to be CBOR primitives, which could mean a good amount of work. An alternative but simpler
        approach is to implement :meth:`to_shallow_primitive` instead. `to_shallow_primitive` allows the returned object
        to be either CBOR :const:`Primitive` or a :class:`CBORSerializable`, as long as the :class:`CBORSerializable`
        does not refer to itself, which could cause infinite loops.
    """

    def to_shallow_primitive(self) -> Union[Primitive, CBORSerializable]:
        """
        Convert the instance to a CBOR primitive. If the primitive is a container, e.g. list, dict, the type of
        its elements could be either a Primitive or a CBORSerializable.

        Returns:
            :const:`Primitive`: A CBOR primitive.

        Raises:
            SerializeException: When the object could not be converted to CBOR primitive
                types.
        """
        raise NotImplementedError(
            f"'to_shallow_primitive()' is not implemented by {self.__class__}."
        )

    def to_primitive(self) -> Primitive:
        """Convert the instance and its elements to CBOR primitives recursively.

        Returns:
            :const:`Primitive`: A CBOR primitive.

        Raises:
            SerializeException: When the object or its elements could not be converted to
                CBOR primitive types.
        """
        result = self.to_shallow_primitive()

        def _dfs(value, freeze=False):
            if isinstance(value, CBORSerializable):
                return _dfs(value.to_primitive(), freeze)
            elif isinstance(value, (dict, OrderedDict, defaultdict)):
                _dict = type(value)()
                if hasattr(value, "default_factory"):
                    _dict.setdefault(value.default_factory)
                for k, v in value.items():
                    _dict[_dfs(k, freeze=True)] = _dfs(v, freeze)
                if freeze:
                    return FrozenDict(_dict)
                return _dict
            elif isinstance(value, set):
                _set = set(_dfs(v, freeze=True) for v in value)
                if freeze:
                    return frozenset(_set)
                return _set
            elif isinstance(value, tuple):
                return tuple(_dfs(v, freeze) for v in value)
            elif isinstance(
                value, (IndefiniteFrozenList, FrozenList, IndefiniteList, list)
            ):
                _list = [_dfs(v, freeze) for v in value]

                already_frozen = isinstance(value, (IndefiniteFrozenList, FrozenList))
                should_freeze = already_frozen or freeze

                if not should_freeze:
                    return (
                        IndefiniteList(_list)
                        if isinstance(value, IndefiniteList)
                        else _list
                    )

                is_indefinite = isinstance(
                    value, (IndefiniteFrozenList, IndefiniteList)
                )
                fl = IndefiniteFrozenList(_list) if is_indefinite else FrozenList(_list)
                fl.freeze()
                return fl
            elif isinstance(value, CBORTag):
                return CBORTag(value.tag, _dfs(value.value, freeze))
            else:
                return value

        return _dfs(result)

    def validate(self):
        """Validate the data stored in the current instance. Defaults to always pass.

        Raises:
            InvalidDataException: When the data is invalid.
        """
        type_hints = get_type_hints(self.__class__)

        def _check_recursive(value, type_hint):
            if type_hint is Any:
                return True

            if isinstance(value, CBORSerializable):
                value.validate()

            origin = getattr(type_hint, "__origin__", None)
            if origin is None:
                return isinstance(value, type_hint)
            elif origin is ClassVar:
                return _check_recursive(value, type_hint.__args__[0])
            elif origin is Union:
                return any(_check_recursive(value, arg) for arg in type_hint.__args__)
            elif origin is Dict or isinstance(value, (dict, FrozenDict)):
                key_type, value_type = type_hint.__args__
                return all(
                    _check_recursive(k, key_type) and _check_recursive(v, value_type)
                    for k, v in value.items()
                )
            elif origin in (list, set, tuple, frozenset, OrderedSet):
                if value is None:
                    return True
                args = type_hint.__args__
                if len(args) == 1:
                    return all(_check_recursive(item, args[0]) for item in value)
                elif len(args) > 1:
                    return all(
                        _check_recursive(item, arg) for item, arg in zip(value, args)
                    )
            return True  # We don't know how to check this type

        for field_name, field_type in type_hints.items():
            field_value = getattr(self, field_name)
            if not _check_recursive(field_value, field_type):
                raise TypeError(
                    f"Field '{field_name}' should be of type {field_type}, "
                    f"got {repr(field_value)} instead."
                )

    def to_validated_primitive(self) -> Primitive:
        """Convert the instance and its elements to CBOR primitives recursively with data validated by :meth:`validate`
        method.

        Returns:
            :const:`Primitive`: A CBOR primitive.

        Raises:
            SerializeException: When the object or its elements could not be converted to
                CBOR primitive types.
        """
        self.validate()
        return self.to_primitive()

    @classmethod
    def from_primitive(
        cls: Type[CBORBase], value: Any, type_args: Optional[tuple] = None
    ) -> CBORBase:
        """Turn a CBOR primitive to its original class type.

        Args:
            cls (CBORBase): The original class type.
            value (:const:`Primitive`): A CBOR primitive.
            type_args (Optional[tuple]): Type arguments for the class.

        Returns:
            CBORBase: A CBOR serializable object.

        Raises:
            DeserializeException: When the object could not be restored from primitives.
        """
        raise NotImplementedError(
            f"'from_primitive()' is not implemented by {cls.__name__}."
        )

    def to_cbor(self) -> bytes:
        """Encode a Python object into CBOR bytes.

        Returns:
            bytes: Python object encoded in cbor bytes.

        Examples:
            >>> class Test(CBORSerializable):
            ...     def __init__(self, number1, number2):
            ...         self.number1 = number1
            ...         self.number2 = number2
            ...
            ...     def to_primitive(value):
            ...         return [value.number1, value.number2]
            ...
            ...     @classmethod
            ...     def from_primitive(cls, value):
            ...         return cls(value[0], value[1])
            ...
            ...     def __repr__(self):
            ...         return f"Test({self.number1}, {self.number2})"
            >>> a = Test(1, 2)
            >>> a.to_cbor().hex()
            '820102'
        """
        return dumps(self, default=default_encoder)

    def to_cbor_hex(self) -> str:
        """Encode a Python object into CBOR hex.

        Returns:
            str: Python object encoded in cbor hex string.
        """
        return self.to_cbor().hex()

    @classmethod
    def from_cbor(cls: Type[CBORBase], payload: Union[str, bytes]) -> CBORBase:
        """Restore a CBORSerializable object from a CBOR.

        Args:
            payload (Union[str, bytes]): CBOR bytes or hex string to restore from.

        Returns:
            CBORBase: Restored CBORSerializable object of the specific subclass type.

        Examples:

            Basic use case:

            >>> class Test(CBORSerializable):
            ...     def __init__(self, number1, number2):
            ...         self.number1 = number1
            ...         self.number2 = number2
            ...
            ...     def to_primitive(value):
            ...         return [value.number1, value.number2]
            ...
            ...     @classmethod
            ...     def from_primitive(cls, value):
            ...         return cls(value[0], value[1])
            ...
            ...     def __repr__(self):
            ...         return f"Test({self.number1}, {self.number2})"
            >>> a = Test(1, 2)
            >>> cbor_hex = a.to_cbor_hex()
            >>> print(Test.from_cbor(cbor_hex))
            Test(1, 2)

            For a CBORSerializable that has CBORSerializables as attributes, we will need to pass
            each child value to the :meth:`from_primitive` method of its corresponding CBORSerializable. Example:

            >>> class TestParent(CBORSerializable):
            ...     def __init__(self, number1, test):
            ...         self.number1 = number1
            ...         self.test = test
            ...
            ...     def to_shallow_primitive(value): # Implementing `to_shallow_primitive` simplifies the work.
            ...         return [value.number1, value.test]
            ...
            ...     @classmethod
            ...     def from_primitive(cls, value):
            ...         test = Test.from_primitive(value[1]) # Restore test by passing `value[1]` to
            ...                                              # `Test.from_primitive`
            ...         return cls(value[0], test)
            ...
            ...     def __repr__(self):
            ...         return f"TestParent({self.number1}, {self.test})"
            >>> a = Test(1, 2)
            >>> b = TestParent(3, a)
            >>> b
            TestParent(3, Test(1, 2))
            >>> cbor_hex = b.to_cbor_hex()
            >>> cbor_hex
            '8203820102'
            >>> print(TestParent.from_cbor(cbor_hex))
            TestParent(3, Test(1, 2))

        """
        if type(payload) is str:
            payload = bytes.fromhex(payload)

        assert isinstance(payload, bytes)

        value = cbor2.loads(payload)

        return cls.from_primitive(value)

    def __repr__(self):
        return pformat(vars(self), indent=2)

    @property
    def json_type(self) -> str:
        """
        Return the class name of the CBORSerializable object.

        This property provides a default string representing the type of the object for use in JSON serialization.

        Returns:
            str: The class name of the object.
        """
        return self.__class__.__name__

    @property
    def json_description(self) -> str:
        """
        Return the docstring of the CBORSerializable object's class.

        This property provides a default string description of the object for use in JSON serialization.

        Returns:
            str: The docstring of the object's class.
        """
        return self.__class__.__doc__ or "Generated with PyCardano"

    def to_json(
        self,
        key_type: Optional[str] = None,
        description: Optional[str] = None,
        **kwargs,
    ) -> str:
        """
        Convert the CBORSerializable object to a JSON string containing type, description, and CBOR hex.

        This method returns a JSON representation of the object, including its type, description, and CBOR hex encoding.

        Args:
            key_type (str): The type to use in the JSON output. Defaults to the class name.
            description (str): The description to use in the JSON output. Defaults to the class docstring.
            **kwargs: Extra key word arguments to be passed to `json.dumps()`

        Returns:
            str: The JSON string representation of the object.
        """
        if "indent" not in kwargs:
            kwargs["indent"] = 2

        return json.dumps(
            {
                "type": key_type or self.json_type,
                "description": description or self.json_description,
                "cborHex": self.to_cbor_hex(),
            },
            **kwargs,
        )

    @classmethod
    def from_json(cls: Type[CBORSerializable], data: str) -> CBORSerializable:
        """
        Load a CBORSerializable object from a JSON string containing its CBOR hex representation.

        Args:
            data (str): The JSON string to load the object from.

        Returns:
            CBORSerializable: The loaded CBORSerializable object.

        Raises:
            DeserializeException: If the loaded object is not of the expected type.
        """
        obj = json.loads(data)

        k = cls.from_cbor(obj["cborHex"])

        if not isinstance(k, cls):
            raise DeserializeException(
                f"Expected type {cls.__name__} but got {type(k).__name__}."
            )

        return k

    def save(
        self,
        path: str,
        key_type: Optional[str] = None,
        description: Optional[str] = None,
        **kwargs,
    ):
        """
        Save the CBORSerializable object to a file in JSON format.

        This method writes the object's JSON representation to the specified file path.
         It raises an error if the file already exists and is not empty.

        Args:
            path (str): The file path to save the object to.
            key_type (str, optional): The type to use in the JSON output. Defaults to the class name.
            description (str, optional): The description to use in the JSON output. Defaults to the class docstring.
            **kwargs: Extra key word arguments to be passed to `json.dumps()`

        Raises:
            IOError: If the file already exists and is not empty.
        """
        if os.path.isfile(path) and os.stat(path).st_size > 0:
            raise IOError(f"File {path} already exists!")
        with open(path, "w") as f:
            f.write(self.to_json(key_type=key_type, description=description, **kwargs))

    @classmethod
    def load(cls, path: str):
        """
        Load a CBORSerializable object from a file containing its JSON representation.

        Args:
            path (str): The file path to load the object from.

        Returns:
            CBORSerializable: The loaded CBORSerializable object.
        """
        with open(path) as f:
            return cls.from_json(f.read())


def _restore_dataclass_field(
    f: Field, v: Primitive
) -> Union[Primitive, CBORSerializable]:
    """Try to restore a value back to its original type based on information given in field.

    Args:
        f (dataclass_field): A data class field.
        v (:const:`Primitive`): A CBOR primitive.

    Returns:
        Union[:const:`Primitive`, CBORSerializable]: A CBOR primitive or a CBORSerializable.
    """

    if "object_hook" in f.metadata:
        return f.metadata["object_hook"](v)
    return _restore_typed_primitive(cast(Any, f.type), v)


def _restore_typed_primitive(
    t: typing.Type, v: Primitive
) -> Union[Primitive, CBORSerializable]:
    """Try to restore a value back to its original type based on information given in field.

    Args:
        f (type): A type
        v (:const:`Primitive`): A CBOR primitive.

    Returns:
        Union[:const:`Primitive`, CBORSerializable]: A CBOR primitive or a CBORSerializable.
    """

    is_cbor_serializable = False
    try:
        is_cbor_serializable = issubclass(t, CBORSerializable)
    except TypeError:
        # Handle the case when t is a generic alias
        origin = typing.get_origin(t)
        if origin is not None:
            try:
                is_cbor_serializable = issubclass(origin, CBORSerializable)
            except TypeError:
                pass

    if t is Any or (t in PRIMITIVE_TYPES and isinstance(v, t)):
        return v
    elif is_cbor_serializable:
        if "type_args" in getfullargspec(t.from_primitive).args:
            args = typing.get_args(t)
            return t.from_primitive(v, type_args=args)
        else:
            return t.from_primitive(v)
    elif hasattr(t, "__origin__") and (t.__origin__ is list):
        t_args = t.__args__
        if len(t_args) != 1:
            raise DeserializeException(
                f"List types need exactly one type argument, but got {t_args}"
            )
        t_subtype = t_args[0]
        if not isinstance(v, (list, IndefiniteList)):
            raise DeserializeException(f"Expected type list but got {type(v)}")
        v_list = [_restore_typed_primitive(t_subtype, w) for w in v]
        return v.__class__(v_list)
    elif isclass(t) and t == ByteString:
        if not isinstance(v, bytes):
            raise DeserializeException(f"Expected type bytes but got {type(v)}")
        return ByteString(v)
    elif hasattr(t, "__origin__") and (t.__origin__ is dict):
        t_args = t.__args__
        if len(t_args) != 2:
            raise DeserializeException(
                f"Dict types need exactly two type arguments, but got {t_args}"
            )
        key_t = t_args[0]
        val_t = t_args[1]
        if not isinstance(v, dict):
            raise DeserializeException(f"Expected dict type but got {type(v)}")
        return {
            _restore_typed_primitive(key_t, key): _restore_typed_primitive(val_t, val)
            for key, val in v.items()
        }
    elif hasattr(t, "__origin__") and (
        t.__origin__ is Union or t.__origin__ is Optional
    ):
        t_args = t.__args__
        for t in t_args:
            try:
                return _restore_typed_primitive(t, v)
            except DeserializeException:
                pass
        raise DeserializeException(
            f"Cannot deserialize object: \n{v}\n in any valid type from {t_args}."
        )
    elif isclass(t) and issubclass(t, IndefiniteList):
        try:
            return t(v)
        except TypeError:
            raise DeserializeException(f"Can not initialize IndefiniteList from {v}")
    raise DeserializeException(f"Cannot deserialize object: \n{v}\n to type {t}.")


ArrayBase = TypeVar("ArrayBase", bound="ArrayCBORSerializable")
"""A generic type that is bounded by ArrayCBORSerializable."""


@dataclass(repr=False)
class ArrayCBORSerializable(CBORSerializable):
    """
    A base class that can serialize its child `dataclass <https://docs.python.org/3/library/dataclasses.html>`_
    into a `CBOR array <https://datatracker.ietf.org/doc/html/rfc8610#section-3.4>`_.

    The class is useful when the position of each item in a list have its own semantic meaning.

    Examples:

        Basic usages:

        >>> from dataclasses import dataclass
        >>> @dataclass
        ... class Test1(ArrayCBORSerializable):
        ...     a: str
        ...     b: str=None
        >>> @dataclass
        ... class Test2(ArrayCBORSerializable):
        ...     c: str
        ...     test1: Test1
        >>> t = Test2(c="c", test1=Test1(a="a"))
        >>> t
        Test2(c='c', test1=Test1(a='a', b=None))
        >>> cbor_hex = t.to_cbor_hex() # doctest: +SKIP
        >>> cbor_hex # doctest: +SKIP
        '826163826161f6'
        >>> Test2.from_cbor(cbor_hex) # doctest: +SKIP
        Test2(c='c', test1=Test1(a='a', b=None))

        A value of `None` will be encoded as nil (#7.22) in cbor. This will become a problem if the field is meant to be
        optional. To exclude an optional attribute from cbor, we can use `field` constructor with a metadata field
        "optional" set to True and default value set to `None`.

        .. Note::
            In ArrayCBORSerializable, all non-optional fields have to be declared before any optional field.

        Example:

        >>> from dataclasses import dataclass, field
        >>> @dataclass
        ... class Test1(ArrayCBORSerializable):
        ...     a: str
        ...     b: str=field(default=None, metadata={"optional": True})
        >>> @dataclass
        ... class Test2(ArrayCBORSerializable):
        ...     c: str
        ...     test1: Test1
        >>> t = Test2(c="c", test1=Test1(a="a"))
        >>> t
        Test2(c='c', test1=Test1(a='a', b=None))
        >>> t.to_primitive() # Notice below that attribute "b" is not included in converted primitive.
        ['c', ['a']]
        >>> cbor_hex = t.to_cbor_hex() # doctest: +SKIP
        >>> cbor_hex # doctest: +SKIP
        '826163816161'
        >>> Test2.from_cbor(cbor_hex) # doctest: +SKIP
        Test2(c='c', test1=Test1(a='a', b=None))
    """

    def to_shallow_primitive(self) -> Primitive:
        """
        Returns:
            :const:`Primitive`: A CBOR primitive.

        Raises:
            SerializeException: When the object could not be converted to CBOR primitive
                types.
        """
        primitives = []
        for f in fields(self):
            val = getattr(self, f.name)
            if val is None and f.metadata.get("optional"):
                continue
            primitives.append(val)
        return primitives

    @classmethod
    @limit_primitive_type(list, tuple, IndefiniteList)
    def from_primitive(
        cls: Type[ArrayBase], values: Union[list, tuple, IndefiniteList]
    ) -> ArrayBase:
        """Restore a primitive value to its original class type.

        Args:
            cls (ArrayBase): The original class type.
            values (List[Primitive]): A list whose elements are CBOR primitives.

        Returns:
            :const:`ArrayBase`: Restored object.

        Raises:
            DeserializeException: When the object could not be restored from primitives.
        """
        all_fields = [f for f in fields(cls) if f.init]

        restored_vals = []
        type_hints = get_type_hints(cls)
        for f, v in zip(all_fields, values):
            if not isclass(f.type):
                f.type = type_hints[f.name]
            v = _restore_dataclass_field(f, v)
            restored_vals.append(v)
        obj = cls(*restored_vals)
        for i in range(len(all_fields), len(values)):
            setattr(obj, f"unknown_field{i - len(all_fields)}", values[i])
        return obj

    def __repr__(self):
        return super().__repr__()


MapBase = TypeVar("MapBase", bound="MapCBORSerializable")
"""A generic type that is bounded by MapCBORSerializable."""


@dataclass(repr=False)
class MapCBORSerializable(CBORSerializable):
    """
    A base class that can serialize its child `dataclass <https://docs.python.org/3/library/dataclasses.html>`_
    into a `CBOR Map <https://datatracker.ietf.org/doc/html/rfc8610#section-3.5.1>`_.

    The class is useful when each key in a map have its own semantic meaning.

    Examples:

        Basic usage:

        >>> from dataclasses import dataclass, field
        >>> @dataclass
        ... class Test1(MapCBORSerializable):
        ...     a: str=""
        ...     b: str=""
        >>> @dataclass
        ... class Test2(MapCBORSerializable):
        ...     c: str=None
        ...     test1: Test1=field(default_factory=Test1)
        >>> t = Test2(test1=Test1(a="a"))
        >>> t
        Test2(c=None, test1=Test1(a='a', b=''))
        >>> t.to_primitive()
        {'c': None, 'test1': {'a': 'a', 'b': ''}}
        >>> cbor_hex = t.to_cbor_hex() # doctest: +SKIP
        >>> cbor_hex # doctest: +SKIP
        'a26163f6657465737431a261616161616260'
        >>> Test2.from_cbor(cbor_hex) # doctest: +SKIP
        Test2(c=None, test1=Test1(a='a', b=''))

        In the example above, all keys in the map share the same name as their corresponding attributes. However,
        sometimes we want to use different keys when serializing some attributes, this could be achieved by adding a
        "key" value to the metadata of a field. Example:

        >>> from dataclasses import dataclass, field
        >>> @dataclass
        ... class Test1(MapCBORSerializable):
        ...     a: str=field(default="", metadata={"key": "0"})
        ...     b: str=field(default="", metadata={"key": "1"})
        >>> @dataclass
        ... class Test2(MapCBORSerializable):
        ...     c: str=field(default=None, metadata={"key": "0", "optional": True})
        ...     test1: Test1=field(default_factory=Test1, metadata={"key": "1"})
        >>> t = Test2(test1=Test1(a="a"))
        >>> t
        Test2(c=None, test1=Test1(a='a', b=''))
        >>> t.to_primitive()
        {'1': {'0': 'a', '1': ''}}
        >>> cbor_hex = t.to_cbor_hex() # doctest: +SKIP
        >>> cbor_hex # doctest: +SKIP
        'a16131a261306161613160'
        >>> Test2.from_cbor(cbor_hex) # doctest: +SKIP
        Test2(c=None, test1=Test1(a='a', b=''))
    """

    def to_shallow_primitive(self) -> Primitive:
        primitives = {}
        for f in fields(self):
            if "key" in f.metadata:
                key = f.metadata["key"]
            else:
                key = f.name
            if key in primitives:
                raise SerializeException(f"Key: '{key}' already exists in the map.")
            val = getattr(self, f.name)
            if val is None and f.metadata.get("optional"):
                continue
            primitives[key] = val
        return primitives

    @classmethod
    @limit_primitive_type(dict, FrozenDict)
    def from_primitive(cls: Type[MapBase], values: Union[dict, FrozenDict]) -> MapBase:
        """Restore a primitive value to its original class type.

        Args:
            cls (MapBase): The original class type.
            values (:const:`Primitive`): A CBOR primitive.

        Returns:
            :const:`MapBase`: Restored object.

        Raises:
            :class:`pycardano.exception.DeserializeException`: When the object could not be restored from primitives.
        """
        all_fields = {f.metadata.get("key", f.name): f for f in fields(cls) if f.init}

        kwargs = {}
        type_hints = get_type_hints(cls)
        for key in values:
            if key not in all_fields:
                raise DeserializeException(f"Unexpected map key {key} in CBOR.")
            f = all_fields[key]
            v = values[key]
            if not isclass(f.type):
                f.type = type_hints[f.name]
            v = _restore_dataclass_field(f, v)
            kwargs[f.name] = v
        return cls(**kwargs)

    def __repr__(self):
        return super().__repr__()


DictBase = TypeVar("DictBase", bound="DictCBORSerializable")
"""A generic type that is bounded by DictCBORSerializable."""


class DictCBORSerializable(CBORSerializable):
    """A dictionary class where all keys share the same type and all values share the same type.

    Examples:

        >>> @dataclass
        ... class Test1(ArrayCBORSerializable):
        ...     a: int
        ...     b: str
        >>>
        >>> class Test2(DictCBORSerializable):
        ...     KEY_TYPE = str
        ...     VALUE_TYPE = Test1
        >>>
        >>> t = Test2()
        >>> t["x"] = Test1(a=1, b="x")
        >>> t["y"] = Test1(a=2, b="y")
        >>> primitives = t.to_primitive()
        >>> deserialized = Test2.from_primitive(primitives)
        >>> assert t == deserialized
        >>> t[1] = 2
        Traceback (most recent call last):
         ...
        typeguard.TypeCheckError: int is not an instance of str
    """

    KEY_TYPE = Type[Any]
    VALUE_TYPE = Type[Any]

    def __init__(self, *args, **kwargs):
        self.data = dict(*args, **kwargs)

    def __getattr__(self, item):
        return getattr(self.data, item)

    def __setitem__(self, key: Any, value: Any):
        check_type(key, self.KEY_TYPE)
        check_type(value, self.VALUE_TYPE)
        self.data[key] = value

    def __getitem__(self, key):
        return self.data[key]

    def __eq__(self, other):
        if isinstance(other, DictCBORSerializable):
            return self.data == other.data
        else:
            return False

    def __len__(self):
        return len(self.data)

    def __iter__(self):
        return iter(self.data)

    def __delitem__(self, key):
        del self.data[key]

    def __repr__(self):
        return self.data.__repr__()

    def __copy__(self):
        return self.__class__(self)

    def __deepcopy__(self, memo):
        return self.__class__(deepcopy(self.data, memo))

    def validate(self):
        for key, value in self.data.items():
            if isinstance(key, CBORSerializable):
                key.validate()
            if isinstance(value, CBORSerializable):
                value.validate()

    def to_shallow_primitive(self) -> dict:
        # Sort keys in a map according to https://datatracker.ietf.org/doc/html/rfc7049#section-3.9
        def _get_sortable_val(key):
            if isinstance(key, CBORSerializable):
                cbor_bytes = key.to_cbor()
            else:
                cbor_bytes = dumps(key)
            return len(cbor_bytes), cbor_bytes

        return dict(sorted(self.data.items(), key=lambda x: _get_sortable_val(x[0])))

    @classmethod
    @limit_primitive_type(dict)
    def from_primitive(cls: Type[DictBase], value: dict) -> DictBase:
        """Restore a primitive value to its original class type.

        Args:
            cls (DictBase): The original class type.
            value (:const:`Primitive`): A CBOR primitive.

        Returns:
            :const:`DictBase`: Restored object.

        Raises:
            DeserializeException: When the object could not be restored from primitives.
        """
        restored = cls()
        for k, v in value.items():
            k = (
                cls.KEY_TYPE.from_primitive(k)  # type: ignore
                if isclass(cls.KEY_TYPE) and issubclass(cls.KEY_TYPE, CBORSerializable)
                else k
            )
            v = (
                cls.VALUE_TYPE.from_primitive(v)  # type: ignore
                if isclass(cls.VALUE_TYPE)
                and issubclass(cls.VALUE_TYPE, CBORSerializable)
                else v
            )
            restored[k] = v
        return restored

    def copy(self) -> DictCBORSerializable:
        return self.__class__(self)


@typechecked
def list_hook(
    cls: Type[CBORBase],
) -> Callable[[List[Primitive]], List[CBORBase]]:
    """A factory that generates a Callable which turns a list of Primitive to a list of CBORSerializables.

    Args:
        cls (CBORBase): The type of CBORSerializable the list will be converted to.

    Returns:
        Callable[[List[Primitive]], List[CBORBase]]: An Callable that restores a list of Primitive to a list of
            CBORSerializables.
    """
    return lambda vals: [cls.from_primitive(v) for v in vals]


class OrderedSet(Generic[T], CBORSerializable):
    def __init__(
        self,
        iterable: Optional[Union[List[T], IndefiniteList]] = None,
        use_tag: bool = True,
    ):
        super().__init__()
        self._dict: Dict[bytes, int] = {}
        self._list: List[T] = []
        self._use_tag = use_tag
        self._is_indefinite_list = False
        if iterable:
            self._is_indefinite_list = isinstance(iterable, IndefiniteList)
            self.extend(iterable)

    def append(self, item: T) -> None:
        if item in self:
            return
        self._list.append(item)
        self._dict[dumps(item, default=default_encoder)] = len(self._list) - 1

    def extend(self, items: Iterable[T]) -> None:
        self._is_indefinite_list = isinstance(items, IndefiniteList)
        for item in items:
            self.append(item)

    def remove(self, item: T) -> None:
        if item not in self:
            return
        index = self._dict.pop(dumps(item, default=default_encoder))
        self._list.pop(index)
        # Update the indices in the dictionary
        for key, idx in self._dict.items():
            if idx > index:
                self._dict[key] = idx - 1

    def __contains__(self, item: object) -> bool:
        return dumps(item, default=default_encoder) in self._dict

    def __iter__(self):
        return iter(self._list)

    def __getitem__(self, index: int) -> T:
        return self._list[index]

    def __len__(self) -> int:
        return len(self._list)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, OrderedSet):
            if isinstance(other, list):
                return list(self) == other
            return False
        return list(self) == list(other)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({list(self)})"

    def to_shallow_primitive(
        self,
    ) -> Union[CBORTag, List[T], IndefiniteList, FrozenList, IndefiniteFrozenList]:
        fields: Union[IndefiniteFrozenList, FrozenList]
        if self._is_indefinite_list:
            fields = IndefiniteFrozenList(list(self))
        else:
            fields = FrozenList(self)
        fields.freeze()
        if self._use_tag:
            return CBORTag(
                258,
                fields,
            )
        return fields

    @classmethod
    def from_primitive(
        cls: Type[OrderedSet[T]], value: Primitive, type_args: Optional[tuple] = None
    ) -> OrderedSet[T]:
        assert (
            type_args is None or len(type_args) == 1
        ), "OrderedSet should have exactly one type argument"
        # Retrieve the type arguments from the class
        type_arg = type_args[0] if type_args else None

        if isinstance(value, CBORTag) and value.tag == 258:
            if isclass(type_arg) and issubclass(type_arg, CBORSerializable):
                value.value = [type_arg.from_primitive(v) for v in value.value]
            return cls(value.value, use_tag=True)

        use_tag = isinstance(value, set)

        if isinstance(value, (list, tuple, set)):
            if isclass(type_arg) and issubclass(type_arg, CBORSerializable):
                value = [type_arg.from_primitive(v) for v in value]

            # If the value is a set, we know it is coming from a CBORTag (#6.258)
            return cls(list(value), use_tag=use_tag)

        raise ValueError(f"Cannot deserialize {value} to {cls}")

    def __deepcopy__(self, memo):
        return self.__class__(deepcopy(list(self), memo), use_tag=self._use_tag)

    def __hash__(self):
        return hash(self.to_shallow_primitive())


class NonEmptyOrderedSet(OrderedSet[T]):
    def __init__(
        self,
        iterable: Optional[Union[List[T], IndefiniteList]] = None,
        use_tag: bool = True,
    ):
        super().__init__(iterable, use_tag)

    def validate(self):
        if not self:
            raise ValueError("NonEmptyOrderedSet cannot be empty")

    @classmethod
    def from_primitive(
        cls: Type[NonEmptyOrderedSet[T]],
        value: Primitive,
        type_args: Optional[tuple] = None,
    ) -> NonEmptyOrderedSet[T]:
        result = cast(NonEmptyOrderedSet[T], super().from_primitive(value, type_args))
        if not result:
            raise ValueError("NonEmptyOrderedSet cannot be empty")
        return result


@dataclass(repr=False)
class CodedSerializable(ArrayCBORSerializable):
    """A base class for CBORSerializable types that have a specific code.

    This class provides a mechanism to validate the type of the object based on its first element.

    Examples:
        >>> from dataclasses import dataclass, field
        >>> @dataclass
        ... class TestCoded(CodedSerializable):
        ...     _CODE = 1
        ...     value: str
        >>>
        >>> # Create and serialize an instance
        >>> test = TestCoded("hello")
        >>> primitives = test.to_primitive()
        >>> primitives
        [1, 'hello']
        >>>
        >>> # Deserialize valid data
        >>> restored = TestCoded.from_primitive(primitives)
        >>> restored.value
        'hello'
        >>>
        >>> # Attempting to deserialize with wrong code raises exception
        >>> invalid_data = [2, "hello"]
        >>> TestCoded.from_primitive(invalid_data)  # doctest: +IGNORE_EXCEPTION_DETAIL
        Traceback (most recent call last):
            ...
        DeserializeException: Invalid TestCoded type 2
    """

    _CODE: int = field(init=False)

    @classmethod
    @limit_primitive_type(list, tuple)
    def from_primitive(
        cls: Type[CodedSerializable], values: Union[list, tuple]
    ) -> CodedSerializable:
        if values[0] != cls._CODE:
            raise DeserializeException(f"Invalid {cls.__name__} type {values[0]}")
        # Cast using Type[CodedSerializable] instead of cls directly
        return cast(Type[CodedSerializable], super()).from_primitive(values[1:])



from typing import Union

from nacl.encoding import RawEncoder
from nacl.hash import blake2b

from pycardano.crypto.bech32 import encode
from pycardano.hash import ScriptHash
from pycardano.transaction import AssetName


def encode_asset(
    policy_id: Union[ScriptHash, bytes, str], asset_name: Union[AssetName, bytes, str]
) -> str:
    """Implementation of CIP14 asset fingerprinting

    This function encodes the asset policy and name into an asset fingerprint, which is
    bech32 compliant.

    For more information:
    https://developers.cardano.org/docs/governance/cardano-improvement-proposals/cip-0014/

    Args:
        policy_id: The asset policy as `ScriptHash`, `bytes`, or a hex `str`
        asset_name: The asset name as `AssetName`, `bytes`, or a hex `str`
    """
    if isinstance(policy_id, str):
        policy_id = bytes.fromhex(policy_id)
    elif isinstance(policy_id, ScriptHash):
        policy_id = policy_id.payload

    if isinstance(asset_name, str):
        asset_name = bytes.fromhex(asset_name)
    elif isinstance(asset_name, AssetName):
        asset_name = asset_name.payload

    asset_hash = blake2b(
        policy_id + asset_name,
        digest_size=20,
        encoder=RawEncoder,
    )

    return encode("asset", asset_hash)
